
/* copyright 1991-96 UserLand Software, Inc. All Rights Reserved.*/


#include <iac.h>
#include "iowacore.h"
#include "iowascript.h"
#include "iowafrontier.h"


/*a little code ghetto for stuff that depends on Frontier or Runtime*/

#ifdef macBirdPlugin
	#ifdef fldebug
		#define alertdialog(s)		DebugStr (s)
	#else
		#define alertdialog(s)		((void *)0)
	#endif
#endif

#define idfrontier 			'LAND'

#define startcardtoken 			'init'
#define setcardtoken 			'sett'
#define closecardtoken 			'clos'
#define edittabletoken 			'edtb'
#define getttabletoken 			'gett'
#define doscripttoken			'dosc' /*DW 9/28/95*/
#define addtoglossarytoken 		'addg' /*DW 10/9/95*/
#define editglossarytoken		'edtg' /*DW 10/9/95*/
#define refglossarytoken		'refg' /*DW 10/9/95*/
#define editobjecttoken 		'edto' /*DW 10/17/95*/
#define runuserscripttoken		'russ' /*DW 11/4/95*/
#define runlinkfiltertoken		'rulf' /*DW 11/4/95*/
#define copytopagetabletoken 	'ctpt' /*DW 11/5/95*/
#define copyfrompagetabletoken 	'cfpt' /*DW 11/5/95*/
#define runpagefiltertoken		'rupf' /*DW 11/5/95*/
#define runfinalfiltertoken		'ruff' /*DW 12/7/95*/


static bigstring lasttablename = "\p";


#define scriptmodebits kOSANullMode




static void getappspec (FSSpec *fs) {
	
	ProcessSerialNumber psn;
	ProcessInfoRec info;
	Str255 s;
	
	GetCurrentProcess (&psn);
	
	info.processInfoLength = (long) sizeof (info);
	
	info.processName = s; /*place to store process name*/
	
	info.processAppSpec = fs; /*place to store process filespec*/
	
	GetProcessInformation (&psn, &info);
	} /*getappspec*/
	
	
static void componenterrormessage (OSErr ec, bigstring errormessage) {

	copystring ("\pComponent Manager error. Its code number is ", errormessage);
	
	pushlong (ec, errormessage);
	
	pushstring ("\p.", errormessage);
	} /*componenterrormessage*/

	
static boolean getcompiledscript (OSAID *scriptid, ComponentInstance *scriptcomp, bigstring errormessage) {
	
	static boolean flinitialized = false;
	static Handle hcompiledscript = nil;
	static ComponentInstance comp = 0;
	static OSAID id = 0;
	
	if (!flinitialized) { /*first time we're called, load the script from resource fork*/
		
		if (hcompiledscript == nil) {
		
			hcompiledscript = GetResource ('scpt', 128);
		
			if (hcompiledscript == nil) {
			
				copystring ("\pThere's no 'scpt' 128 resource available", errormessage);
			
				return (false);
				}
			}
		
		if (comp == 0) {
		
			comp = findlangcomponent (idfrontier);
		
			if (comp == 0)  {
			
				copystring ("\pFrontier 3.0 or greater isn't running.", errormessage);
			
				return (false);
				}
			}
		
		if (id == 0) {
		
			AEDesc scriptdesc;
			OSErr ec;
			
			scriptdesc.descriptorType = /*typeOSAGenericStorage*/ 'scpt';
	
			scriptdesc.dataHandle = hcompiledscript; 
	
			ec = OSALoad (comp, &scriptdesc, kOSANullMode, &id);
	
			if (ec != noErr) {
			
				componenterrormessage (ec, errormessage);
			
				return (false);
				}
			}
		
		flinitialized = true;
		}
	
	*scriptcomp = comp;
	
	*scriptid = id;
	
	return (true);
	} /*getcompiledscript*/

	
static boolean callcompiledscript (AppleEvent *message, OSType resulttype, bigstring errormessage, Handle *hreturned) {
	
	ComponentInstance scriptcomp;
	OSAID scriptid, resultid;
	AEDesc resultdesc = {typeNull, nil};
	OSErr ec;

	setstringlength (errormessage, 0);
	
	*hreturned = nil;
	
	if (!getcompiledscript (&scriptid, &scriptcomp, errormessage))
		return (false);
	
	ec = OSAExecuteEvent (scriptcomp, message, scriptid, scriptmodebits, &resultid);
	
	if (ec == noErr) {
		
		ec = OSACoerceToDesc (scriptcomp, resultid, resulttype, kOSANullMode, &resultdesc);
		
		OSADispose (scriptcomp, resultid);
		}
	
	/*DW 11/17/95 -- site of a former memory leak*/ {
		
		AEDisposeDesc (message);	
		}
	
	switch (ec) {
		
		case noErr: 
			copyhandle (resultdesc.dataHandle, hreturned); /*script ran without error*/
			
			break;
		
		case errOSAScriptError: 
			if (OSAScriptError (scriptcomp, kOSAErrorMessage, typeChar, &resultdesc) == noErr)
				texthandletostring (resultdesc.dataHandle, errormessage);
			else
				copystring ("\pThe script generated an error, but no message was provided.", errormessage);
			
			break;
		
		default: 
			componenterrormessage (ec, errormessage);
			
			break;
		} /*switch*/
	
	AEDisposeDesc (&resultdesc);
	
	return (*hreturned != nil);
	} /*callcompiledscript*/
	
	
static boolean newevent (OSType token, AppleEvent *event) {

	AEAddressDesc selfAddress; 
	ProcessSerialNumber psn;
	OSErr ec;

	psn.highLongOfPSN = 0;

	psn.lowLongOfPSN = kCurrentProcess;

	ec = AECreateDesc (typeProcessSerialNumber, (Ptr) &psn, sizeof (psn), &selfAddress);
	
	if (ec != noErr)
		return (false);
	
	ec = AECreateAppleEvent ('abcd', token, &selfAddress, kAutoGenerateReturnID, kAnyTransactionID, event);
	
	AEDisposeDesc (&selfAddress);
	
	return (ec == noErr);
	} /*newevent*/


boolean frontStartCard (AEDesc *filelist) {
	
	/*
	a card is starting up. we call the script at system.verbs.traps.IRUN.stcd
	to create a table for it, unpack the packed table at that location, and 
	copy in the filespec for the application and any files that were dragged
	onto the app.
	
	if tablename is empty, the script assigns a name and returns it. we return
	it to the caller, who is expected to link it into iowadata.
	*/
	
	bigstring tablename;
	boolean flcopyname;
	FSSpec fsapp;
	AppleEvent event;
	
	getappspec (&fsapp);
		
	/*pull the table name from the tablename handle*/ {
	
		if ((**iowadata).tablename == nil) { /*let the script assign a name to the table*/
			
			setstringlength (tablename, 0);
			
			flcopyname = true;
			}
		else {
			texthandletostring ((**iowadata).tablename, tablename);
			
			flcopyname = false;
			}
		}
	
	if (!newevent (startcardtoken, &event))
		return (false);
		
	IACglobals.event = &event;
	
	if (!IACpushstringparam (tablename, 'prm1'))
		goto error;
	
	if (!IACpushbinaryparam ((**iowadata).embeddedtable, 'tabl', 'prm2'))
		goto error;
	
	if (!IACpushfilespecparam (&fsapp, 'prm3'))
		goto error;
	
	if (filelist == nil) {
		
		AEDesc desc;
		
		desc.descriptorType = typeNull;
		
		desc.dataHandle = nil;
		
		if (AEPutParamDesc (IACglobals.event, 'prm4', &desc) != noErr)
			goto error;
		}
	else
		if (AEPutParamDesc (IACglobals.event, 'prm4', filelist) != noErr)
			goto error;
			
	/*send the event to the compiled script*/ {
		
		hdlcard savediowadata = iowadata; /*if startCard script calls a cardEditor verb, iowadata will be smashed*/
		bigstring errormessage;
		Handle htablename;
		
		if (!callcompiledscript (&event, typeChar, errormessage, &htablename)) {
			
			iowadata = savediowadata;
			
			alertdialog (errormessage);
			
			return (false);
			}
			
		iowadata = savediowadata;
		
		(**iowadata).tablename = htablename;
			
		texthandletostring (htablename, lasttablename);
		}
		
	return (true);
	
	error:
	
	AEDisposeDesc (&event);	
	
	return (false);
	} /*frontStartCard*/
	

boolean frontSetRuntimeCard (boolean flmajorswitch) {
	
	/*
	iowadata contains a card that's running. call the script at
	system.verbs.traps.IRUN.sett to make it the current card, 
	all the items in its table are global. and if it's a major
	switch, it's menubar is installed at system.menubars.DRPb.
	*/

	AppleEvent event;
	bigstring tablename;
	
	if ((**iowadata).notscriptedcard)
		return (true);
		
	texthandletostring ((**iowadata).tablename, tablename);
	
	if (equalstrings (tablename, lasttablename)) /*minimize AE traffic*/
		return (true);
		
	copystring (tablename, lasttablename);
	
	if (!newevent (setcardtoken, &event))
		goto error;
		
	IACglobals.event = &event;
	
	if (!IACpushstringparam (tablename, 'prm1'))
		goto error;
	
	if (!IACpushbooleanparam (flmajorswitch, 'prm2'))
		goto error;
	
	/*send the event to the compiled script*/ {
		
		bigstring errormessage;
		Handle hresult;
		
		if (!callcompiledscript (&event, typeChar, errormessage, &hresult)) {
			
			alertdialog (errormessage);
			
			return (false);
			}
			
		disposehandle (hresult);
		}
		
	return (true);
	
	error:
	
	AEDisposeDesc (&event);	
	
	return (false);
	} /*frontSetRuntimeCard*/
	
	
boolean frontEditTable (bigstring tablewindowtitle) {
	
	AppleEvent event;
	bigstring bs;
	
	if ((**iowadata).notscriptedcard)
		return (true);
		
	if (!newevent (edittabletoken, &event))
		goto error;
		
	IACglobals.event = &event;
	
	texthandletostring ((**iowadata).tablename, bs);
	
	if (!IACpushstringparam (bs, 'prm1'))
		goto error;
	
	if (!IACpushstringparam (tablewindowtitle, 'prm2'))
		goto error;
	
	/*send the event to the compiled script*/ {
		
		bigstring errormessage;
		Handle hresult;
		
		if (!callcompiledscript (&event, typeChar, errormessage, &hresult)) {
			
			alertdialog (errormessage);
			
			return (false);
			}
			
		disposehandle (hresult);
		}
		
	return (true);
	
	error:
	
	AEDisposeDesc (&event);	
	
	return (false);
	} /*frontEditTable*/
	
	
boolean frontGetTable (void) { 
	
	/*
	send a message to Frontier requesting the table associated with
	the card in iowadata.
	*/
	
	AppleEvent event;
	bigstring bs;
	
	if ((**iowadata).notscriptedcard)
		return (true);
		
	if (!newevent (getttabletoken, &event))
		return (false);
		
	IACglobals.event = &event;
	
	texthandletostring ((**iowadata).tablename, bs);
	
	if (!IACpushstringparam (bs, 'prm1'))
		goto error;
		
	/*send the event to the compiled script*/ {
		
		bigstring errormessage;
		Handle hpackedtable;
		
		if (!callcompiledscript (&event, 'tabl', errormessage, &hpackedtable)) {
			
			alertdialog (errormessage);
			
			return (false);
			}
		
		assert (hpackedtable != nil);
	
		(**iowadata).embeddedtable = hpackedtable;
		}
		
	return (true);
	
	error:
	
	AEDisposeDesc (&event);	
	
	return (false);
	} /*frontGetTable*/
	

boolean frontDeleteCardTable (void) {
	
	/*
	iowadata contains a running card that is being terminated. we call
	system.verbs.traps.IRUN.delt to delete the table for this card.
	*/

	AppleEvent event;
	bigstring bs;
	
	if ((**iowadata).notscriptedcard) /*this card is handled by C code*/
		return (true);
	
	if (!newevent (closecardtoken, &event))
		goto error;
		
	IACglobals.event = &event;
	
	texthandletostring ((**iowadata).tablename, bs);
	
	if (!IACpushstringparam (bs, 'prm1'))
		goto error;
	
	/*send the event to the compiled script*/ {
		
		bigstring errormessage;
		Handle hresult;
		
		if (!callcompiledscript (&event, typeChar, errormessage, &hresult)) {
			
			alertdialog (errormessage);
			
			return (false);
			}
			
		disposehandle (hresult);
		}
		
	return (true);
	
	error:
	
	AEDisposeDesc (&event);	
	
	return (false);
	} /*frontDeleteCardTable*/
	
	
boolean frontRunMacro (bigstring macro, bigstring errorstring, Handle *hresult) {
	
	AppleEvent event;
	
	if (!newevent (doscripttoken, &event))
		goto error;
		
	IACglobals.event = &event;
	
	if (!IACpushstringparam (macro, 'prm1'))
		goto error;
	
	if (!callcompiledscript (&event, typeChar, errorstring, hresult))
		return (false);
		
	return (true);
	
	error:
	
	AEDisposeDesc (&event);	
	
	return (false);
	} /*frontRunMacro*/
	
	
boolean frontRunUserScript (bigstring macro, bigstring path, bigstring errorstring, Handle *hresult) {
	
	AppleEvent event;
	
	if (!newevent (runuserscripttoken, &event))
		goto error;
		
	IACglobals.event = &event;
	
	if (!IACpushstringparam (macro, 'prm1'))
		goto error;
	
	if (!IACpushstringparam (path, 'prm2'))
		goto error;
	
	if (!callcompiledscript (&event, typeChar, errorstring, hresult))
		return (false);
		
	return (true);
	
	error:
	
	AEDisposeDesc (&event);	
	
	return (false);
	} /*frontRunUserScript*/
	
	
boolean frontRunLinkFilter (Handle hlinkstring, short linktype, bigstring errorstring, Handle *hresult) {
	
	AppleEvent event;
	
	if (!newevent (runlinkfiltertoken, &event))
		return (false);
		
	IACglobals.event = &event;
	
	if (!IACpushtextparam (hlinkstring, 'prm1'))
		goto error;
	
	if (!IACpushshortparam (linktype, 'prm2'))
		goto error;
	
	if (!callcompiledscript (&event, typeChar, errorstring, hresult))
		return (false);
		
	return (true);
	
	error:
	
	AEDisposeDesc (&event);	
	
	return (false);
	} /*frontRunLinkFilter*/
	
	
boolean frontRunPageFilter (bigstring errorstring) {
	
	AppleEvent event;
	Handle hresult;
	
	setstringlength (errorstring, 0);
	
	if (!newevent (runpagefiltertoken, &event))
		return (false);
		
	if (!callcompiledscript (&event, typeChar, errorstring, &hresult))
		goto error;
	
	disposehandle (hresult);
	
	return (true);
	
	error:
	
	AEDisposeDesc (&event);	
	
	return (false);
	} /*frontRunPageFilter*/
	
	
boolean frontRunFinalFilter (Handle htext, bigstring errorstring, Handle *hresult) {
	
	AppleEvent event;
	
	setstringlength (errorstring, 0);
	
	if (!newevent (runfinalfiltertoken, &event))
		return (false);
		
	IACglobals.event = &event;
	
	if (!IACpushtextparam (htext, 'prm1'))
		goto error;
	
	if (!callcompiledscript (&event, typeChar, errorstring, hresult))
		goto error;
	
	return (true);
	
	error:
	
	AEDisposeDesc (&event);	
	
	return (false);
	} /*frontRunFinalFilter*/
	
	
boolean frontCopyToPageTable (AERecord *rec) {
	
	AppleEvent event;
	bigstring errorstring;
	Handle hresult;
	
	if (!newevent (copytopagetabletoken, &event))
		goto error;
		
	IACglobals.event = &event;
	
	if (!IACpushrecordparam (rec, 'prm1'))
		goto error;
	
	if (!callcompiledscript (&event, typeChar, errorstring, &hresult))
		return (false);
	
	disposehandle (hresult);
	
	return (true);
	
	error:
	
	AEDisposeDesc (&event);	
	
	return (false);
	} /*frontCopyToPageTable*/
	
	
boolean frontCopyFromPageTable (AERecord *rec) {

	AppleEvent event;
	bigstring errorstring;
	Handle hresult;
	
	if (!newevent (copyfrompagetabletoken, &event))
		return (false);
		
	if (!callcompiledscript (&event, typeAERecord, errorstring, &hresult))
		goto error;
		
	(*rec).descriptorType = typeAERecord;
	
	(*rec).dataHandle = hresult;
		
	return (true);
	
	error:
	
	AEDisposeDesc (&event);	
	
	return (false);
	} /*frontCopyFromPageTable*/
	
	
boolean frontAddToGlossary (void *name, bigstring value, short type, bigstring errorstring) {
	
	AppleEvent event;
	Handle hresult = nil;
	
	if (!newevent (addtoglossarytoken, &event))
		goto error;
		
	IACglobals.event = &event;
	
	if (!IACpushstringparam (name, 'prm1'))
		goto error;
	
	if (!IACpushstringparam (value, 'prm2'))
		goto error;
	
	if (!IACpushshortparam (type, 'prm3'))
		goto error;
	
	if (!callcompiledscript (&event, typeChar, errorstring, &hresult))
		return (false);
		
	disposehandle (hresult);

	return (true);
	
	error:
	
	AEDisposeDesc (&event);	
	
	disposehandle (hresult);
	
	return (false);
	} /*frontAddToGlossary*/


boolean frontEditGlossary (bigstring errorstring) {
	
	AppleEvent event;
	Handle hresult = nil;
	
	if (!newevent (editglossarytoken, &event))
		goto error;
		
	IACglobals.event = &event;
	
	if (!callcompiledscript (&event, typeChar, errorstring, &hresult))
		return (false);
		
	disposehandle (hresult);

	return (true);
	
	error:
	
	AEDisposeDesc (&event);	
	
	disposehandle (hresult);
	
	return (false);
	} /*frontEditGlossary*/


boolean frontGlossaryReference (bigstring name, bigstring errorstring, Handle *hresult) {
	
	AppleEvent event;
	
	*hresult = nil;
	
	if (!newevent (refglossarytoken, &event))
		return (false);
		
	IACglobals.event = &event;
	
	if (!IACpushstringparam (name, 'prm1'))
		goto error;
	
	if (!callcompiledscript (&event, typeChar, errorstring, hresult))
		return (false);
		
	return (true);
	
	error:
	
	AEDisposeDesc (&event);	
	
	disposehandle (*hresult);
	
	return (false);
	} /*frontGlossaryReference*/


boolean frontEditObject (bigstring celladdress, bigstring errorstring) {
	
	AppleEvent event;
	Handle hresult = nil;
	
	if (!newevent (editobjecttoken, &event))
		return (false);
		
	IACglobals.event = &event;
	
	if (!IACpushstringparam (celladdress, 'prm1'))
		goto error;
	
	if (!callcompiledscript (&event, typeChar, errorstring, &hresult))
		return (false);
		
	disposehandle (hresult);
	
	return (true);
	
	error:
	
	AEDisposeDesc (&event);	
	
	disposehandle (hresult);
	
	return (false);
	} /*frontEditObject*/


